
/**
 * Verwaltet das Gewölbe
 * 
 * @author Albert Wiedemann 
 * @version 1.0
 */
class Gewoelbe
{
    Raum ausgang;
    boolean [] markierungen;
    Spiel spiel;

    /**
     * Legt die Grundstrukturen an
     * @param spiel das Spiel, zu dem das Gewölbe gehört
     */
    Gewoelbe (Spiel spiel)
    {
        ausgang = null;
        markierungen = new boolean [Konstanten. anzahlRäume];
        this.spiel = spiel;
    }

    /**
     * Untersucht den Raum und seine Nachbarn
     * @param akt der aktuelle Raum
     * @param nummer die Nummer des gesuchten Raums
     * @returen den gefundenen Raum oder null
     */
    Raum RaumLokalSuchen (Raum akt, int nummer)
    {
        if (!markierungen [akt.nummer])
        {
            markierungen [akt.nummer] = true;
            if (akt. nummer == nummer)
            {
                return akt;
            }
            else
            {
                for (Raum wohin: akt.türen.türen)
                {
                    Raum hi = RaumLokalSuchen (wohin, nummer);
                    if (hi != null)
                    {
                        return hi;
                    }
                }
                return null;
            }
        }
        return null;
    }

    /**
     * Sucht den Raum mit der angegebenen Nummer
     */
    Raum RaumSuchen (int nummer)
    {
        for (int i = 0; i < Konstanten.anzahlRäume; i += 1)
        {
            markierungen [i] = false;
        }
        return RaumLokalSuchen (ausgang, nummer);
    }

    /**
     * Trägt die Daten für die Bildschirmdarstellung der Landspiel.anzeige ein.
     * @param alle wenn wahr werden alle Räume gzeichnet, sonst nur die bekannten
     * @param markierung der Raum, der markiert werden soll (i.d.R. der Raum des Spielers)
     */
    void AnzeigeZeichnen (boolean alle, Raum markierung)
    {
        Raum akt;
        int mode;
        spiel.anzeige.LandkartendatenLöschen();
        for (int i = 0; i < Konstanten.anzahlRäume; i += 1)
        {
            akt = RaumSuchen(i);
            if (alle || akt. betreten)
            {
                if (akt == markierung)
                {
                    mode = Konstanten.markiert;
                }
                else if (!alle && akt.betreten)
                {
                    mode = Konstanten.besucht;
                    if (akt.ausgangGesucht)
                    {
                        mode = Konstanten.ausgangGesucht;
                    }
                }
                else
                {
                    mode = Konstanten.nichtBesucht;
                }
                spiel.anzeige.RaumEintragen(new LandkarteneintragRaum(akt.nummer, akt.x, akt.y, mode));
                for (Raum wohin: akt.türen.türen)
                {
                    spiel.anzeige.WegEintragen(new LandkarteneintragWeg(akt.x, akt.y, wohin.x, wohin.y));
                    if (!alle && !wohin.betreten)
                    {
                        spiel.anzeige.RaumEintragen(new LandkarteneintragRaum (wohin.nummer, wohin.x, wohin.y, Konstanten.nichtBesucht));
                    }
                }
            }
        }
        spiel.anzeige.LandkarteNeuZeichnen();
    }

    /**
     * Trägt einen neuen Raum in das Raster ein, falls der Platz frei ist und Weg vom aktuellen Raum dorthin möglich ist.
     * @param von der aktuelle Raum
     * @param x1 die x-Koordinaten des gewünschten Zielraums
     * @param y1 die y-Koordinaten des gewünschten Zielraums
     * @param nummer die Nummer des gewünschten Raums
     * @param r das Testraster für die Landspiel.anzeigendarstellung
     */
    boolean RaumEintragen (Raum von, int x1, int y1, int nummer, Raster r)
    {
        Raum nach;
        if (r.RaumOkTesten(x1, y1) && r.WegTesten (x1, y1, von.x, von.y))
        {
            r.RaumSetzen (x1, y1);
            r.WegZeichnen (x1, y1, von.x, von.y);
            nach = new Raum (nummer, x1, y1);
            nach.NeueTürEintragen(von);
            von.NeueTürEintragen(nach);
            return true;
        }
        return false;
    }
    
    /**
     * Einen neuen Raum an den bestehenden Wegbaum anfügen
     * @param nummer die Nummer des gewünschten Raums
     * @param r das Testraster für die Landspiel.anzeigendarstellung
     */
    void RaumAnfügen (int nummer, Raster r)
    {
        Raum von;
        int zähler;
        int radius;
        zähler = 0;
        for (int i = 1; i <= Konstanten.anzahlRäume; i += 1)
        {
            do
            {
                von = RaumSuchen(Zufall.zufall.GanzzahlGeben(nummer));
                radius = 1;
                switch (Zufall.zufall.GanzzahlGeben(10))
                {
                    case 9:
                    case 8:
                        radius ++;
                    case 7:
                    case 6:
                    case 5:
                        radius ++;
                }
                while (radius > 0)
                {
                    switch (radius)
                    {
                        case 1:
                            for (int x1 = von.x - 1; x1 <= von.x + 1; x1 += 1)
                            {
                                for (int y1 = von.y - 1; y1 <= von.y + 1; y1 += 1)
                                {
                                    if ((x1 != von.x) || (y1 != von.y))
                                    {
                                        if (RaumEintragen (von, x1, y1, nummer, r))
                                        {
                                            return;
                                        }
                                    }
                                }
                            }
                            break;
                        case 2:
                            for (int x1 = von.x - 2; x1 <= von.x + 2; x1 += 1)
                            {
                                if (RaumEintragen (von, x1, von. y - 2, nummer, r))
                                {
                                    return;
                                }
                            }
                            for (int x1 = von. x - 2; x1 <= von. x + 2; x1 += 1)
                            {
                                if (RaumEintragen (von, x1, von. y + 2, nummer, r))
                                {
                                    return;
                                }
                            }
                            for (int y1 = von. y - 1; y1 <= von. y + 1; y1 += 1)
                            {
                                if (RaumEintragen (von, von. x - 2, y1, nummer, r))
                                {
                                    return;
                                }
                            }
                            for (int y1 = von. y - 1; y1 <= von. y + 1; y1 += 1)
                            {
                                if (RaumEintragen (von, von. x + 2, y1, nummer, r))
                                {
                                    return;
                                }
                            }
                            break;
                        case 3:
                            for (int x1 = von.x - 3; x1 <= von.x + 3; x1 += 1)
                            {
                                if (RaumEintragen (von, x1, von.y - 3, nummer, r))
                                {
                                    return;
                                }
                            }
                            for (int x1 = von.x - 3; x1 <= von.x + 3; x1 += 1)
                            {
                                if (RaumEintragen (von, x1, von. y + 3, nummer, r))
                                {
                                    return;
                                }
                            }
                            for (int y1 = von.y - 2; y1 <= von.y + 2; y1 += 1)
                            {
                                if (RaumEintragen (von, von. x - 3, y1, nummer, r))
                                {
                                    return;
                                }
                            }
                            for (int y1 = von.y - 2; y1 <= von.y + 2; y1 += 1)
                            {
                                if (RaumEintragen (von, von. x + 3, y1, nummer, r))
                                {
                                    return;
                                }
                            }
                            break;
                    }
                    radius -= 1;
                }
                zähler += 1;
            }
            while (zähler < Konstanten.anzahlRäume * 2);
        }
        zähler = 0;
        zähler = 5 / zähler;
    }
    
    /**
     * Erzeugt einen Weg zwischen zwei zufälligen Räumen
     * @param r das Testraster für die Landspiel.anzeigendarstellung
     */
    void ZufallsWegErzeugen (Raster r)
    {
        final int maxRadius2 = Konstanten.breiteRaumRaster * Konstanten.breiteRaumRaster +
                                Konstanten.höheRaumRaster  * Konstanten.höheRaumRaster;
        int radius, dx, dy;
        Raum von, nach;
        von = RaumSuchen (Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
        radius = 0;
        do
        {
            radius += 1;
            for (int nummer = 1; nummer <= Konstanten.anzahlRäume; nummer += 1)
            {
                nach = RaumSuchen(Zufall.zufall.GanzzahlGeben(nummer));
                dx = nach.x - von.x;
                dy = nach.y - von.y;
                dx = dx * dx;
                dy = dy * dy;
                if (((radius - 1) * (radius - 1) < dx + dy) && (radius * radius >= dx + dy) &&
                    r.WegTesten(von.x, von.y, nach.x, nach.y))
                {
                    r.WegZeichnen(von.x, von.y, nach.x, nach.y);
                    nach.NeueTürEintragen(von);
                    von.NeueTürEintragen(nach);
                    return;
                }
            }
        }
        while (radius * radius < maxRadius2);
    }

    /**
     * Tauscht die Raumnummer zufällig durch
     */
    void NummernTauschen ()
    {
        int hilf;
        Raum von, nach;
        for (int i = 1; i <= Konstanten.anzahlRäume * 2; i += 1)
        {
            von = RaumSuchen(Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
            nach = RaumSuchen(Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
            hilf = nach.nummer;
            nach.nummer = von.nummer;
            von.nummer = hilf;
        }
    }

    /**
     * erzeugt alle Schätze im Gewölbe
     */
    void SchätzeGenerieren ()
    {
        Raum akt;
        for (int i = 1; i <= 3; i++)
        {
            do
            {
                akt = RaumSuchen(Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
            }
            while (!akt.schätze.isEmpty());
            akt.schätze.add(new Schluessel(spiel));
        }
        for (int i = 1; i <= Konstanten.anzahlRäume * 3 / 4; i++)
        {
            do
            {
                akt = RaumSuchen(Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
            }
            while (!akt.schätze.isEmpty());
            akt.schätze.add(Schatz.ZufälligGenerieren(spiel));
        }
    }
    
    /**
     * erzeugt alle Monster im Gewölbe
     */
    void MonsterGenerieren ()
    {
        Raum akt;
        for (int i = 1; i <= Konstanten.anzahlRäume * 2 / 3; i += 1)
        {
            do
            {
                akt = RaumSuchen(Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
            }
            while (akt.monster != null);
            akt.monster = Monster.ZufälligGenerieren(spiel);
        }
        for (int i = 0; i < Konstanten.anzahlRäume; i += 1)
        {
            akt = RaumSuchen(i);
            if (akt. monster == null)
            {
                akt.monster = new LeeresMonster (spiel);
            }
        }
    }
    
    /**
     * Erzeugt eine neues Gewölbe
     */
    void Generieren ()
    {
        Raster r = new Raster ();
        ausgang = new Raum (0, Zufall.zufall.GanzzahlGeben(Konstanten.breiteRaumRaster), Zufall.zufall.GanzzahlGeben(Konstanten.höheRaumRaster));
        r.RaumSetzen(ausgang.x, ausgang.y);
        for (int i = 1; i < Konstanten.anzahlRäume; i += 1)
        {
            RaumAnfügen(i, r);
        }
        for (int i = 0; i < Konstanten.anzahlRäume; i += 1)
        {
            ZufallsWegErzeugen(r);
        }
        NummernTauschen();
        SchätzeGenerieren();
        MonsterGenerieren();
    }

    /**
     * Gibt einen zufällig gewählten Startraum zurück
     * @return Startraum
     */
    Raum StartRaum ()
    {
        return RaumSuchen(Zufall.zufall.GanzzahlGeben(Konstanten.anzahlRäume));
    }
    
    /**
     * Testet, ob ein Raum der Ausgang ist
     * @param test der zu testende Raum
     * @return wahr, wenn der Raum der Ausgang ist.
     */
    boolean AusgangTesten (Raum test)
    {
        return ausgang == test;
    }
}
